@@ -18,7 +18,9 @@ use crate::bbox::BoundingBox;
 use crate::borrow_element_as;
 use crate::css::{self, Origin, Stylesheet};
 use crate::dpi::Dpi;
-use crate::drawing_ctx::{draw_tree, with_saved_cr, DrawingMode, SvgNesting};
+use crate::drawing_ctx::{
+    draw_tree, with_saved_cr, DrawingMode, RenderingConfiguration, SvgNesting,
+};
 use crate::error::{AcquireError, InternalRenderingError, LoadingError, NodeIdError};
 use crate::io::{self, BinaryData};
 use crate::is_element_of_type;
@@ -123,6 +125,32 @@ impl LoadOptions {
     }
 }
 
+/// Document-level rendering options.
+///
+/// This gets then converted to a [`drawing_ctx::RenderingConfiguration`] when all the
+/// parameters are known.
+pub struct RenderingOptions {
+    pub dpi: Dpi,
+    pub cancellable: Option<gio::Cancellable>,
+    pub user_language: UserLanguage,
+    pub svg_nesting: SvgNesting,
+    pub testing: bool,
+}
+
+impl RenderingOptions {
+    /// Copies the options to a [`RenderingConfiguration`], and adds the `measuring` flag.
+    fn to_rendering_configuration(&self, measuring: bool) -> RenderingConfiguration {
+        RenderingConfiguration {
+            dpi: self.dpi,
+            cancellable: self.cancellable.clone(),
+            user_language: self.user_language.clone(),
+            svg_nesting: self.svg_nesting,
+            testing: self.testing,
+            measuring,
+        }
+    }
+}
+
 /// A loaded SVG file and its derived data.
 pub struct Document {
     /// Tree of nodes; the root is guaranteed to be an `<svg>` element.
@@ -269,22 +297,10 @@ impl Document {
         session: &Session,
         cr: &cairo::Context,
         viewport: &cairo::Rectangle,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        svg_nesting: SvgNesting,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(), InternalRenderingError> {
         let root = self.root();
-        self.render_layer(
-            session,
-            cr,
-            root,
-            viewport,
-            user_language,
-            dpi,
-            svg_nesting,
-            is_testing,
-        )
+        self.render_layer(session, cr, root, viewport, options)
     }
 
     pub fn render_layer(
@@ -293,10 +309,7 @@ impl Document {
         cr: &cairo::Context,
         node: Node,
         viewport: &cairo::Rectangle,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        svg_nesting: SvgNesting,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(), InternalRenderingError> {
         cr.status()?;
 
@@ -304,17 +317,15 @@ impl Document {
 
         let viewport = Rect::from(*viewport);
 
+        let config = options.to_rendering_configuration(false);
+
         with_saved_cr(cr, || {
             draw_tree(
                 session.clone(),
                 DrawingMode::LimitToStack { node, root },
                 cr,
                 viewport,
-                user_language,
-                dpi,
-                svg_nesting,
-                false,
-                is_testing,
+                config,
                 &mut AcquiredNodes::new(self),
             )
             .map(|_bbox| ())
@@ -326,25 +337,21 @@ impl Document {
         session: &Session,
         node: Node,
         viewport: Rect,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(Rect, Rect), InternalRenderingError> {
         let root = self.root();
 
         let target = cairo::ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
         let cr = cairo::Context::new(&target)?;
 
+        let config = options.to_rendering_configuration(true);
+
         let bbox = draw_tree(
             session.clone(),
             DrawingMode::LimitToStack { node, root },
             &cr,
             viewport,
-            user_language,
-            dpi,
-            SvgNesting::Standalone,
-            true,
-            is_testing,
+            config,
             &mut AcquiredNodes::new(self),
         )?;
 
@@ -359,14 +366,11 @@ impl Document {
         session: &Session,
         node: Node,
         viewport: &cairo::Rectangle,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(cairo::Rectangle, cairo::Rectangle), InternalRenderingError> {
         let viewport = Rect::from(*viewport);
 
-        let (ink_rect, logical_rect) =
-            self.geometry_for_layer(session, node, viewport, user_language, dpi, is_testing)?;
+        let (ink_rect, logical_rect) = self.geometry_for_layer(session, node, viewport, options)?;
 
         Ok((
             cairo::Rectangle::from(ink_rect),
@@ -378,25 +382,21 @@ impl Document {
         &self,
         session: &Session,
         node: &Node,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<BoundingBox, InternalRenderingError> {
         let target = cairo::ImageSurface::create(cairo::Format::Rgb24, 1, 1)?;
         let cr = cairo::Context::new(&target)?;
 
         let node = node.clone();
 
+        let config = options.to_rendering_configuration(true);
+
         draw_tree(
             session.clone(),
             DrawingMode::OnlyNode(node),
             &cr,
             unit_rectangle(),
-            user_language,
-            dpi,
-            SvgNesting::Standalone,
-            true,
-            is_testing,
+            config,
             &mut AcquiredNodes::new(self),
         )
     }
@@ -406,11 +406,9 @@ impl Document {
         &self,
         session: &Session,
         node: Node,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(cairo::Rectangle, cairo::Rectangle), InternalRenderingError> {
-        let bbox = self.get_bbox_for_element(session, &node, user_language, dpi, is_testing)?;
+        let bbox = self.get_bbox_for_element(session, &node, options)?;
 
         let ink_rect = bbox.ink_rect.unwrap_or_default();
         let logical_rect = bbox.rect.unwrap_or_default();
@@ -430,13 +428,11 @@ impl Document {
         cr: &cairo::Context,
         node: Node,
         element_viewport: &cairo::Rectangle,
-        user_language: &UserLanguage,
-        dpi: Dpi,
-        is_testing: bool,
+        options: &RenderingOptions,
     ) -> Result<(), InternalRenderingError> {
         cr.status()?;
 
-        let bbox = self.get_bbox_for_element(session, &node, user_language, dpi, is_testing)?;
+        let bbox = self.get_bbox_for_element(session, &node, options)?;
 
         if bbox.ink_rect.is_none() || bbox.rect.is_none() {
             // Nothing to draw
@@ -459,16 +455,14 @@ impl Document {
             cr.scale(factor, factor);
             cr.translate(-ink_r.x0, -ink_r.y0);
 
+            let config = options.to_rendering_configuration(false);
+
             draw_tree(
                 session.clone(),
                 DrawingMode::OnlyNode(node),
                 cr,
                 unit_rectangle(),
-                user_language,
-                dpi,
-                SvgNesting::Standalone,
-                false,
-                is_testing,
+                config,
                 &mut AcquiredNodes::new(self),
             )
             .map(|_bbox| ())
@@ -772,6 +766,7 @@ pub struct AcquiredNodes<'i> {
     document: &'i Document,
     num_elements_acquired: usize,
     node_stack: Rc<RefCell<NodeStack>>,
+    nodes_with_cycles: Vec<Node>,
 }
 
 impl<'i> AcquiredNodes<'i> {
@@ -780,6 +775,7 @@ impl<'i> AcquiredNodes<'i> {
             document,
             num_elements_acquired: 0,
             node_stack: Rc::new(RefCell::new(NodeStack::new())),
+            nodes_with_cycles: Vec::new(),
         }
     }
 
@@ -816,6 +812,10 @@ impl<'i> AcquiredNodes<'i> {
             .lookup_node(node_id)
             .ok_or_else(|| AcquireError::LinkNotFound(node_id.clone()))?;
 
+        if self.nodes_with_cycles.contains(&node) {
+            return Err(AcquireError::CircularReference(node.clone()));
+        }
+
         if node.borrow_element().is_accessed_by_reference() {
             self.acquire_ref(&node)
         } else {
@@ -835,8 +835,11 @@ impl<'i> AcquiredNodes<'i> {
     /// * At the drawing stage, `acquire_ref()` the pattern node that we already had, so that
     /// its child elements that reference other paint servers will be able to detect circular
     /// references to the pattern.
-    pub fn acquire_ref(&self, node: &Node) -> Result<AcquiredNode, AcquireError> {
-        if self.node_stack.borrow().contains(node) {
+    pub fn acquire_ref(&mut self, node: &Node) -> Result<AcquiredNode, AcquireError> {
+        if self.nodes_with_cycles.contains(node) {
+            Err(AcquireError::CircularReference(node.clone()))
+        } else if self.node_stack.borrow().contains(node) {
+            self.nodes_with_cycles.push(node.clone());
             Err(AcquireError::CircularReference(node.clone()))
         } else {
             self.node_stack.borrow_mut().push(node);
